// Copyright 2011 Julian Phillips.  All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package py

// #include "utils.h"
// static inline int moduleCheck(PyObject *o) { return PyModule_Check(o); }
// static inline int moduleCheckE(PyObject *o) { return PyModule_CheckExact(o); }
// static inline void decref(PyObject *obj) { Py_DECREF(obj); }
//import "C"

import "unsafe"

type Module struct {
	AbstractObject
	o _PyObject
}

// ModuleType is the Type object that represents the Module type.
var ModuleType = (*Type)(unsafe.Pointer(&*_PyModule_Type))

func moduleCheck(obj Object) bool {
	if obj == nil {
		return false
	}
	return _moduleCheck(c(obj)) != 0
}

func newModule(obj *_PyObject) *Module {
	return (*Module)(unsafe.Pointer(obj))
}

func Import(name string) (*Module, error) {
	s := _CString(name)
	defer _free(unsafe.Pointer(s))

	pyName := _PyString_FromString(s)
	defer _decref(pyName)

	obj := _PyImport_Import(pyName)
	if obj == nil {
		return nil, exception()
	}

	return newModule(obj), nil
}

func InitModule(name string, methods []Method) (*Module, error) {
	cName := _CString(name)
	defer _free(unsafe.Pointer(cName))

	m := _Py_InitModule4(cName, nil, nil, nil, _PYTHON_API_VERSION)
	if m == nil {
		return nil, exception()
	}

	if len(methods) == 0 {
		return newModule(m), nil
	}

	n := _PyString_FromString(cName)
	if n == nil {
		return nil, exception()
	}

	d := _PyModule_GetDict(m)
	if d == nil {
		return nil, exception()
	}

	for _, method := range methods {
		pyF, err := makeCFunction(method.Name, method.Func, method.Doc, n)
		if err != nil {
			return nil, err
		}

		if _PyDict_SetItemString(d, _CString(method.Name), c(pyF)) != 0 {
			return nil, exception()
		}
	}

	return newModule(m), nil
}

func ExecCodeModule(name string, code Object) (*Module, error) {
	s := _CString(name)
	defer _free(unsafe.Pointer(s))
	ret := _PyImport_ExecCodeModule(s, c(code))
	if ret == nil {
		return nil, exception()
	}
	return newModule(ret), nil
}

func NewModule(name string) (*Module, error) {
	cname := _CString(name)
	defer _free(unsafe.Pointer(cname))

	ret := _PyModule_New(cname)
	if ret == nil {
		return nil, exception()
	}

	return newModule(ret), nil
}

func (mod *Module) CheckExact() bool {
	return _moduleCheckE(c(mod)) != 0
}

func (mod *Module) Dict() *Dict {
	ret := _PyModule_GetDict(c(mod))
	return newDict(ret)
}

func (mod *Module) Name() (string, error) {
	ret := _PyModule_GetName(c(mod))
	if ret == nil {
		return "", exception()
	}
	return _GoString(ret), nil
}

func (mod *Module) Filename() (string, error) {
	ret := _PyModule_GetFilename(c(mod))
	if ret == nil {
		return "", exception()
	}
	return _GoString(ret), nil
}

func (mod *Module) AddObject(name string, obj Object) error {
	if obj == nil {
		return AssertionError.Err("ValueError: obj == nil!")
	}

	cname := _CString(name)
	defer _free(unsafe.Pointer(cname))

	ret := _PyModule_AddObject(c(mod), cname, c(obj))
	if ret < 0 {
		return exception()
	}

	return nil
}

func (mod *Module) AddIntConstant(name string, value int) error {
	cname := _CString(name)
	defer _free(unsafe.Pointer(cname))

	ret := _PyModule_AddIntConstant(c(mod), cname, int64(value))
	if ret < 0 {
		return exception()
	}

	return nil
}

func (mod *Module) AddStringConstant(name, value string) error {
	cname := _CString(name)
	defer _free(unsafe.Pointer(cname))

	cvalue := _CString(value)
	defer _free(unsafe.Pointer(cvalue))

	ret := _PyModule_AddStringConstant(c(mod), cname, cvalue)
	if ret < 0 {
		return exception()
	}

	return nil
}
